% boundary_layer.m
% ad-hoc refinement strategy in 1D
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clc; clear all; close all;

global epsilon
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% initialize parameters

N = 5;

if 0
    % regulare grid
    x_full = linspace(0,1,N+2)';
    x = x_full(2:end-1);
else
    % random grid
    x = sort(rand(N,1));
    x_full = [0; x; 1];
end
    
epsilon = 0.005;

refine_steps = 10;

% show the initial mesh
figure(1);
subplot(2,2,1);
hold on;
P = plot(x_full,0*x_full,'go-',x_full,NaN*x_full,'bx');
T = text(0.1,0.1,'refinement 0');
title('Cells')

subplot(2,2,3);
LP = stairs(x_full,NaN*x_full);
title('L2 (DxxU)');

% place for exact and numerical solutions
subplot(2,2,2);
UP = plot(x_full,0*x_full,'b-');
title('u')

subplot(2,2,4);
DP = semilogy(x_full,0*x_full,'b-'); % difference
title('difference')

x_new = [];

for refine_level = 1:refine_steps

    x_full = sort([x_full;x_new]);
    x = x_full(2:end-1);
    N = length(x);    %subplot(2,1,2);
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % exact solution: for this problem still available
    u_ex = find_ex(x_full);

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % build the FEM matrix
    A = system_matrix(x_full,N);
    
    % rhs
    F = 0.5*(x_full(3:end)-x_full(1:end-2));


    U = A\F;

    U = [0; U; 0]; % add the boundary points
    
    %plot(x_full,u_ex,'ro',x_full,U,'b');


    % estimate dxx U by differences at all inner points
    I = 2:N+1;

    % nodal derivative
    DxxU = 2./(x_full(I+1)-x_full(I-1)).*( (U(I-1)-U(I))./(x_full(I-1)-x_full(I)) - (U(I+1)-U(I))./(x_full(I+1)-x_full(I)));

    DxxU = [DxxU(1); DxxU; DxxU(N)]; % extend to boundary points
    % cell derivative as the mean value of neighbouring nodal derivatives
    DxxU_cell = 0.5*(DxxU(1:end-1) + DxxU(2:end));

    % estimate L2 norm:
    h = (x_full(2:end)-x_full(1:end-1)); % h(i) = x(i+1)-x(i)
    L2_DxxU_cell = sqrt( h.*(DxxU_cell).^2);
    %L2_DxxU_cell = abs(DxxU_cell);

    % indicate cells to be refined
    Normed_L2_DxxU_cell = L2_DxxU_cell./max(abs(L2_DxxU_cell));

    to_refine = find(Normed_L2_DxxU_cell>=0.9);

    x_new = 0.5*(x_full(to_refine)+x_full(to_refine+1));
    
    % show the new points
    set(P(1),'XData',x_full,'YData',0*x_full);
    set(P(2),'XData',x_new,'YData',0*x_new);
    
    set(T,'String',['refinement ',num2str(refine_level)]);
    
    % plot the pw constant values of L2_DxxU_cell (L2 norm at each cell)
    set(LP,'XData',x_full,'YData',[L2_DxxU_cell;L2_DxxU_cell(end)]);
    
    set(UP,'XData',x_full,'YData',U);
    set(DP,'XData',x_full,'YData',abs(U-u_ex));
    
    drawnow;
    waitforbuttonpress
    
    e_adaptive = abs(U-u_ex); % store for later comparison
    x_adaptive = x_full;
end

% plot the final distribution of cell sizes
figure(2)
N = length(x);
plot((1:N+1)',x_full(2:end)-x_full(1:end-1),'r.-');
title('Cell sizes');
xlabel('Cell nr')
ylabel('h')


% now, check what an equidistant grid would do
% first, with the same number of grid points
x_full = (linspace(0,1,N+2))';
A = system_matrix(x_full,N);
% rhs
F = 0.5*(x_full(3:end)-x_full(1:end-2));
U_eq1 = A\F; U_eq1 = [0; U_eq1; 0]; % add the boundary points
u_ex = find_ex(x_full);

e_eq1 = abs(U_eq1-u_ex);
x_eq1 = x_full;



% now, ~the same min cell size
hmin = min(h); 
xt = 0:hmin:1;
N = length(xt);
x_full = (linspace(0,1,N))';
A = system_matrix(x_full,N-2);
% rhs
F = 0.5*(x_full(3:end)-x_full(1:end-2));
U_eq2 = A\F; U_eq2 = [0; U_eq2; 0]; % add the boundary points
u_ex = find_ex(x_full);

e_eq2 = abs(U_eq2-u_ex);
x_eq2 = x_full;

figure(3);
semilogy(x_adaptive,e_adaptive,'r',x_eq1,e_eq1,'b',x_eq2,e_eq2,'g');
xlabel('x');
ylabel('error')
title('Error for different refinements');
legend('adaptive','equidist. with the same N','equidist. with min h',0);


figure(4);
plot(x_adaptive,U,'r-',x_eq1,U_eq1,'b.-');
legend('adaptive','equidistant',0);
disp('N for smallest h:')
disp(N);
disp('N for adaptive')
disp(length(x_eq1));

